@@ -1,25 +1,25 @@
 from base64 import b64encode as encode
+
 import collections.abc
 import datetime
 from functools import partialmethod, total_ordering
 import hashlib
-import operator
 import os
+from typing import TYPE_CHECKING, Any, Callable, Dict, List, Sequence, Tuple, Type, Optional, Union, cast
 import warnings
 
 import pytz
 
 from django import forms
 from django.conf import settings
-from django.core import checks, exceptions, validators
+from django.core import checks, exceptions, validators as dj_validators
 from django.db.models.constants import LOOKUP_SEP
-from django.db.models.fields import BLANK_CHOICE_DASH, NOT_PROVIDED, return_None
+from django.db.models.fields import BLANK_CHOICE_DASH, NOT_PROVIDED, return_None  # type: ignore
 from django.utils import timezone
 from django.utils.dateparse import (
     parse_date,
     parse_datetime
 )
-from django.utils.encoding import smart_text
 from django.utils.functional import cached_property
 from django.utils.text import capfirst
 from django.utils.translation import gettext_lazy as _
@@ -27,6 +27,12 @@ from django.utils.translation import gettext_lazy as _
 from . import forms as ldap_forms
 from .validators import validate_email_forward
 
+if TYPE_CHECKING:
+    from .models import Model
+
+
+Validator = Callable[[Any], None]
+
 
 @total_ordering
 class Field:
@@ -36,12 +42,12 @@ class Field:
     to build Django ORM-like models and fool ModelForm into working for us.
     """
 
-    empty_strings_allowed = True
-    empty_values = list(validators.EMPTY_VALUES)
-    creation_counter = 0
+    empty_strings_allowed: bool = True
+    empty_values: List[Any] = list(dj_validators.EMPTY_VALUES)
+    creation_counter: int = 0
 
-    default_validators = []  # Default set of validators
-    default_error_messages = {
+    default_validators: List[Validator] = []  # Default set of validators
+    default_error_messages: Dict[str, str] = {
         'invalid_choice': _('Value %(value)r is not a valid choice.'),
         'null': _('This field cannot be null.'),
         'blank': _('This field cannot be blank.'),
@@ -49,27 +55,38 @@ class Field:
 
     # These are here to fool ModelForm into thinking we're a Django ORM Field. We
     # don't actually use them.
-    many_to_many = None
-    many_to_one = None
-    one_to_many = None
-    one_to_one = None
-    related_model = None
+    many_to_many: Any = None
+    many_to_one: Any = None
+    one_to_many: Any = None
+    one_to_one: Any = None
+    related_model: Any = None
 
     # Field flags
-    hidden = False
+    hidden: bool = False
 
     # Generic field type description, usually overridden by subclasses
-    def _description(self):
+    def _description(self) -> str:
         return _('Field of type: %(field_type)s') % {
             'field_type': self.__class__.__name__
         }
     description = property(_description)
 
-    def __init__(self, verbose_name=None, name=None, primary_key=False,
-                 max_length=None, blank=False, null=False,
-                 default=NOT_PROVIDED, editable=True, choices=None,
-                 help_text='', validators=(), error_messages=None,
-                 db_column=None):
+    def __init__(
+        self,
+        verbose_name: str = None,
+        name: str = None,
+        primary_key: bool = False,
+        max_length: int = None,
+        blank: bool = False,
+        null: bool = False,
+        default: Any = NOT_PROVIDED,
+        editable: bool = True,
+        choices: List[Any] = None,
+        help_text: str = '',
+        validators: Sequence[Validator] = (),
+        error_messages: Dict[str, str] = None,
+        db_column: str = None
+    ) -> None:
         self.name = name
         self.verbose_name = verbose_name  # May be set by set_attributes_from_name
         self.primary_key = primary_key
@@ -79,25 +96,25 @@ class Field:
         self.editable = editable
         if isinstance(choices, collections.abc.Iterator):
             choices = list(choices)
-        self.choices = choices or []
+        self.choices: List[Any] = choices or []
         self.help_text = help_text
         self.blank = True
         self.db_column = db_column
 
-        self.model = None
+        self.model: Optional[Type["Model"]] = None
 
         self.creation_counter = Field.creation_counter
         Field.creation_counter += 1
 
         self._validators = list(validators)  # Store for deconstruction later
 
-        messages = {}
+        messages: Dict[str, str] = {}
         for c in reversed(self.__class__.__mro__):
             messages.update(getattr(c, 'default_error_messages', {}))
         messages.update(error_messages or {})
         self.error_messages = messages
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         """Display the module, class, and name of the field."""
         path = '%s.%s' % (self.__class__.__module__, self.__class__.__qualname__)
         name = getattr(self, 'name', None)
@@ -105,7 +122,7 @@ class Field:
             return '<%s: %s>' % (path, name)
         return '<%s>' % path
 
-    def __lt__(self, other):
+    def __lt__(self, other: "Field") -> bool:
         """
         We need to implement this because django.forms.models.fields_from_model
         tries to sort all the fields on a model before interrogating them for
@@ -115,23 +132,23 @@ class Field:
             return self.creation_counter < other.creation_counter
         raise NotImplementedError
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         if isinstance(other, Field):
             return self.creation_counter == other.creation_counter
-        return NotImplementedError
+        raise NotImplementedError
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash(self.creation_counter)
 
-    def has_default(self):
+    def has_default(self) -> bool:
         """Return a boolean of whether this field has a default value."""
         return self.default is not NOT_PROVIDED
 
-    def get_default(self):
+    def get_default(self) -> Any:
         """Return the default value for this field."""
         return self._get_default()
 
-    def check(self, **kwargs):
+    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
         return [
             *self._check_field_name(),
             *self._check_choices(),
@@ -139,12 +156,12 @@ class Field:
             *self._check_validators(),
         ]
 
-    def _check_field_name(self):
+    def _check_field_name(self) -> List[checks.Error]:
         """
         Check if field name is valid, i.e. 1) does not end with an
         underscore, 2) does not contain "__" and 3) is not "pk".
         """
-        if self.name.endswith('_'):
+        if cast(str, self.name).endswith('_'):
             return [
                 checks.Error(
                     'Field names must not end with an underscore.',
@@ -152,7 +169,7 @@ class Field:
                     id='fields.E001',
                 )
             ]
-        elif LOOKUP_SEP in self.name:
+        if LOOKUP_SEP in cast(str, self.name):
             return [
                 checks.Error(
                     'Field names must not contain "%s".' % (LOOKUP_SEP,),
@@ -160,7 +177,7 @@ class Field:
                     id='fields.E002',
                 )
             ]
-        elif self.name == 'pk':
+        if self.name == 'pk':
             return [
                 checks.Error(
                     "'pk' is a reserved word that cannot be used as a field name.",
@@ -168,10 +185,9 @@ class Field:
                     id='fields.E003',
                 )
             ]
-        else:
-            return []
+        return []
 
-    def _check_choices(self):
+    def _check_choices(self) -> List[checks.Error]:
         if not self.choices:
             return []
 
@@ -221,7 +237,7 @@ class Field:
             )
         ]
 
-    def _check_null_allowed_for_primary_keys(self):
+    def _check_null_allowed_for_primary_keys(self) -> List[checks.Error]:
         if (self.primary_key and self.null):
             return [
                 checks.Error(
@@ -232,11 +248,10 @@ class Field:
                     id='fields.E007',
                 )
             ]
-        else:
-            return []
+        return []
 
-    def _check_validators(self):
-        errors = []
+    def _check_validators(self) -> List[checks.Error]:
+        errors: List[checks.Error] = []
         for i, validator in enumerate(self.validators):
             if not callable(validator):
                 errors.append(
@@ -255,11 +270,11 @@ class Field:
         return errors
 
     @property
-    def ldap_attribute(self):
-        return self.db_column or self.name
+    def ldap_attribute(self) -> str:
+        return cast(str, self.db_column or self.name)
 
     @cached_property
-    def _get_default(self):
+    def _get_default(self) -> Callable[[], Any]:
         if self.has_default():
             if callable(self.default):
                 return self.default
@@ -269,14 +284,14 @@ class Field:
             return return_None
         return str  # return empty string
 
-    def to_python(self, value):
+    def to_python(self, value: Any) -> Any:
         return value
 
     @cached_property
-    def validators(self):
-        return [*self.default_validators, *self._validators]  # NOQA
+    def validators(self) -> List[Validator]:
+        return [*self.default_validators, *self._validators]
 
-    def run_validators(self, value):
+    def run_validators(self, value: Any) -> None:
         if value in self.empty_values:
             return
 
@@ -292,7 +307,7 @@ class Field:
         if errors:
             raise exceptions.ValidationError(errors)
 
-    def validate(self, value, model_instance):
+    def validate(self, value: Any, model_instance: "Model") -> None:
         """
         Validate value and raise ValidationError if necessary. Subclasses
         should override this to provide validation logic.
@@ -306,7 +321,7 @@ class Field:
                 if isinstance(option_value, (list, tuple)):
                     # This is an optgroup, so look inside the group for
                     # options.
-                    for optgroup_key, optgroup_value in option_value:
+                    for optgroup_key, optgroup_value in option_value:  # pylint: disable=unused-variable
                         if value == optgroup_key:
                             return
                 elif value == option_key:
@@ -323,7 +338,11 @@ class Field:
         if not self.blank and value in self.empty_values:
             raise exceptions.ValidationError(self.error_messages['blank'], code='blank')
 
-    def clean(self, value, model_instance):
+    def pre_save(self, model_instance: "Model", add: bool) -> Any:
+        """Return field's value just before saving."""
+        return getattr(model_instance, self.attname)
+
+    def clean(self, value: Any, model_instance: "Model") -> Any:
         """
         Convert the value's type and run validation. Validation errors
         from to_python() and validate() are propagated. Return the correct
@@ -334,16 +353,24 @@ class Field:
         self.run_validators(value)
         return value
 
-    def set_attributes_from_name(self, name):
+    def set_attributes_from_name(self, name: str) -> None:
         self.name = self.name or name
+        self.attname = self.name
         if self.verbose_name is None and self.name:
             self.verbose_name = self.name.replace('_', ' ')
 
-    def get_choices(self, include_blank=True, blank_choice=BLANK_CHOICE_DASH, limit_choices_to=None):
+    def get_choices(
+        self,
+        include_blank: bool = True,
+        blank_choice: List[Tuple[str, str]] = None,
+        limit_choices_to: Any = None
+    ):
         """
-        Return choices with a default blank choices included, for use
-        as <select> choices for this field.
+        Return choices with a default blank choices included, for use as
+        <select> choices for this field.
         """
+        if not blank_choice:
+            blank_choice = BLANK_CHOICE_DASH
         if self.choices:
             choices = list(self.choices)
             if include_blank:
@@ -351,20 +378,14 @@ class Field:
                 if not blank_defined:
                     choices = blank_choice + choices
             return choices
-        rel_model = self.remote_field.model
-        limit_choices_to = limit_choices_to or self.get_limit_choices_to()
-        choice_func = operator.attrgetter('pk')
-        return (blank_choice if include_blank else []) + [
-            (choice_func(x), smart_text(x))
-            for x in rel_model._default_manager.complex_filter(limit_choices_to)
-        ]
+        return blank_choice
 
     def limit_choices_to(self):
         raise NotImplementedError
 
-    def _get_flatchoices(self):
+    def _get_flatchoices(self) -> List[Tuple[str, str]]:
         """Flattened version of choices tuple."""
-        flat = []
+        flat: List[Tuple[str, str]] = []
         for choice, value in self.choices:
             if isinstance(value, (list, tuple)):
                 flat.extend(value)
@@ -373,11 +394,18 @@ class Field:
         return flat
     flatchoices = property(_get_flatchoices)
 
-    def formfield(self, form_class=None, choices_form_class=None, **kwargs):
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
         """Return a django.forms.Field instance for this field."""
-        defaults = {'required': not self.blank,
-                    'label': capfirst(self.verbose_name),
-                    'help_text': self.help_text}
+        defaults: Dict[str, Any] = {
+            'required': not self.blank,
+            'label': capfirst(self.verbose_name),
+            'help_text': self.help_text
+        }
         if self.has_default():
             if callable(self.default):
                 defaults['initial'] = self.default
@@ -409,10 +437,10 @@ class Field:
             form_class = forms.CharField
         return form_class(**defaults)
 
-    def value_from_object(self, obj):
-        return getattr(obj, self.name)
+    def value_from_object(self, obj: "Model") -> Any:
+        return getattr(obj, cast(str, self.name))
 
-    def from_db_value(self, value):
+    def from_db_value(self, value: List[bytes]) -> Union[List[str], Optional[Any]]:
         """
         Take data for one attribute from LDAP and convert it to our internal
         python format.
@@ -427,7 +455,7 @@ class Field:
         """
         return [b.decode('utf-8') for b in value]
 
-    def to_db_value(self, value):
+    def to_db_value(self, value: Any) -> Dict[str, List[bytes]]:
         # Subclasses should implement this and do proper casting of the value
         # from our internal data type to the appropriate value to stuff into LDAP
         # and then call super().to_db_value(value)
@@ -446,17 +474,17 @@ class Field:
             cleaned.append(item)
         return {self.ldap_attribute: cleaned}
 
-    def value_to_string(self, obj):
+    def value_to_string(self, obj: "Model") -> str:
         return str(self.value_from_object(obj))
 
-    def save_form_data(self, obj, value):
+    def save_form_data(self, obj: "Model", value: Any) -> None:
         """
         This is what django.forms.models.ModelForm calls to save data to our
         LdapModel instance.
         """
-        setattr(obj, self.name, value)
+        setattr(obj, cast(str, self.name), value)
 
-    def contribute_to_class(self, cls, name):
+    def contribute_to_class(self, cls, name: str) -> None:
         """
         Register the field with the model class it belongs to.
         """
@@ -475,23 +503,23 @@ class BooleanField(Field):
     the strings 'true' and 'false' in LDAP.
     """
 
-    empty_strings_allowed = False
+    empty_strings_allowed: bool = False
 
-    default_error_messages = {
+    default_error_messages: Dict[str, str] = {
         'invalid': _("'%(value)s' value must be either True or False."),
         'invalid_nullable': _("'%(value)s' value must be either True, False, or None."),
     }
 
-    description = _("Boolean (Either True or False)")
+    description: str = _("Boolean (Either True or False)")  # type: ignore
 
-    LDAP_TRUE = 'true'
-    LDAP_FALSE = 'false'
+    LDAP_TRUE: str = 'true'
+    LDAP_FALSE: str = 'false'
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
         self.blank = True
 
-    def to_python(self, value):
+    def to_python(self, value: Union[None, bool, str]) -> Optional[bool]:
         if self.null and value in self.empty_values:
             return None
         if value in (True, False):
@@ -508,34 +536,39 @@ class BooleanField(Field):
             params={'value': value},
         )
 
-    def from_db_value(self, value):
-        value = super().from_db_value(value)
-        if value == []:
+    def from_db_value(self, value: List[bytes]) -> Optional[bool]:
+        db_value = cast(List[str], super().from_db_value(value))
+        if db_value == []:
             return None
-        if value[0].lower() == self.LDAP_TRUE:
+        if db_value[0].lower() == self.LDAP_TRUE:
             return True
-        if value[0].lower() == self.LDAP_FALSE:
+        if db_value[0].lower() == self.LDAP_FALSE:
             return False
-        if value == [self.LDAP_TRUE]:
+        if db_value == [self.LDAP_TRUE]:
             return True
-        elif value == [self.LDAP_FALSE]:
+        if db_value == [self.LDAP_FALSE]:
             return False
-        else:
-            raise ValueError('Field "{}" (BooleanField) on model {} got got unexpected data from LDAP: {}'.format(
-                self.name,
-                self.model._meta.object_name,
-                value
-            ))
+        raise ValueError('Field "{}" (BooleanField) on model {} got got unexpected data from LDAP: {}'.format(
+            self.name,
+            self.model._meta.object_name,  # type: ignore
+            db_value
+        ))
 
-    def to_db_value(self, value):
+    def to_db_value(self, value: Optional[bool]) -> Dict[str, List[bytes]]:
+        db_value: Optional[str] = None
         if value is not None:
             if value:
-                value = self.LDAP_TRUE
+                db_value = self.LDAP_TRUE
             else:
-                value = self.LDAP_FALSE
-        return super().to_db_value(value)
+                db_value = self.LDAP_FALSE
+        return super().to_db_value([db_value])
 
-    def formfield(self, **kwargs):
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
         if self.choices:
             include_blank = not (self.has_default() or 'initial' in kwargs)
             defaults = {'choices': self.get_choices(include_blank=include_blank)}
@@ -544,40 +577,52 @@ class BooleanField(Field):
             # In HTML checkboxes, 'required' means "must be checked" which is
             # different from the choices case ("must select some value").
             # required=False allows unchecked checkboxes.
-            defaults = {'form_class': form_class, 'required': False}
+            defaults = {
+                'form_class': form_class,
+                'required': False
+            }
         return super().formfield(**{**defaults, **kwargs})
 
 
 class AllCapsBooleanField(BooleanField):
 
-    LDAP_TRUE = 'TRUE'
-    LDAP_FALSE = 'FALSE'
+    LDAP_TRUE: str = 'TRUE'
+    LDAP_FALSE: str = 'FALSE'
 
 
 class CharField(Field):
 
-    description = _("String (up to %(max_length)s)")
+    description: str = _("String (up to %(max_length)s)")  # type: ignore
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
-        self.validators.append(validators.MaxLengthValidator(self.max_length))
+        self.validators.append(dj_validators.MaxLengthValidator(self.max_length))
 
-    def to_python(self, value):
+    def to_python(self, value: Optional[str]) -> Optional[str]:
         if isinstance(value, str) or value is None:
             return value
         return str(value)
 
-    def from_db_value(self, value):
-        value = super().from_db_value(value)
-        if value == []:
+    def from_db_value(self, value: List[bytes]) -> Optional[str]:
+        db_value = cast(List[str], super().from_db_value(value))
+        if db_value == []:
             return None
-        return value[0]
+        return db_value[0]
 
-    def formfield(self, **kwargs):
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
         # Passing max_length to forms.CharField means that the value's length
         # will be validated twice. This is considered acceptable since we want
         # the value in the form field (to pass into widget for example).
-        defaults = {'max_length': self.max_length}
+        defaults: Dict[str, Any] = {
+            'form_class': form_class,
+            'choices_form_class': choices_form_class,
+            'max_length': self.max_length
+        }
         if self.null:
             defaults['empty_value'] = None
         defaults.update(kwargs)
@@ -586,32 +631,38 @@ class CharField(Field):
 
 class DateField(Field):
 
-    empty_strings_allowed = False
-    default_error_messages = {
+    empty_strings_allowed: bool = False
+    default_error_messages: Dict[str, str] = {
         'invalid': _("'%(value)s' value has an invalid date format. It must be "
                      "in YYYY-MM-DD format."),
         'invalid_date': _("'%(value)s' value has the correct format (YYYY-MM-DD) "
                           "but it is an invalid date."),
     }
-    description = _("Date (without time)")
+    description: str = _("Date (without time)")  # type: ignore
 
-    LDAP_DATETIME_FORMAT = "%Y%m%d"
+    LDAP_DATETIME_FORMAT: str = "%Y%m%d"
 
-    def __init__(self, verbose_name=None, name=None, auto_now=False,
-                 auto_now_add=False, **kwargs):
+    def __init__(
+        self,
+        verbose_name: str = None,
+        name: str = None,
+        auto_now: bool = False,
+        auto_now_add: bool = False,
+        **kwargs
+    ):
         self.auto_now, self.auto_now_add = auto_now, auto_now_add
         if auto_now or auto_now_add:
             kwargs['editable'] = False
             kwargs['blank'] = True
         super().__init__(verbose_name, name, **kwargs)
 
-    def check(self, **kwargs):
+    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
         return [
             *super().check(),
             *self._check_fix_default_value(),
         ]
 
-    def _check_fix_default_value(self):
+    def _check_fix_default_value(self) -> List[checks.Warning]:
         """
         Warn that using an actual date or datetime value is probably wrong;
         it's only evaluated on server startup.
@@ -651,7 +702,7 @@ class DateField(Field):
 
         return []
 
-    def to_python(self, value):
+    def to_python(self, value: Optional[Union[str, datetime.date, datetime.datetime]]) -> Optional[datetime.date]:
         if value is None:
             return value
         if isinstance(value, datetime.datetime):
@@ -681,15 +732,14 @@ class DateField(Field):
             params={'value': value},
         )
 
-    def pre_save(self, model_instance, add):
+    def pre_save(self, model_instance: "Model", add: bool) -> Optional[datetime.date]:
         if self.auto_now or (self.auto_now_add and add):
             value = datetime.date.today()
             setattr(model_instance, self.attname, value)
             return value
-        else:
-            return super().pre_save(model_instance, add)
+        return super().pre_save(model_instance, add)
 
-    def contribute_to_class(self, cls, name, **kwargs):
+    def contribute_to_class(self, cls, name: str, **kwargs) -> None:
         super().contribute_to_class(cls, name, **kwargs)
         if not self.null:
             setattr(
@@ -701,40 +751,49 @@ class DateField(Field):
                 partialmethod(cls._get_next_or_previous_by_FIELD, field=self, is_next=False)
             )
 
-    def from_db_value(self, value):
-        value = super().from_db_value(value)
-        if value is []:
+    def from_db_value(self, value: List[bytes]) -> Optional[datetime.date]:
+        db_value = super().from_db_value(value)
+        if not db_value:
             return None
-        value = value[0]
-        ts = datetime.datetime.strptime(value, self.LDAP_DATETIME_FORMAT)
+        dt = db_value[0]
+        ts = datetime.datetime.strptime(dt, self.LDAP_DATETIME_FORMAT)
         return datetime.date(year=ts.year, month=ts.month, day=ts.day)
 
-    def to_db_value(self, value):
+    def to_db_value(self, value: Optional[Union[datetime.date, datetime.datetime]]) -> Dict[str, List[bytes]]:
+        db_value = None
         if value:
-            value = value.strftime(self.LDAP_DATETIME_FORMAT)
-        return super().to_db_value(value)
+            db_value = value.strftime(self.LDAP_DATETIME_FORMAT)
+        return super().to_db_value(db_value)
 
-    def value_to_string(self, obj):
+    def value_to_string(self, obj: "Model") -> str:
         val = self.value_from_object(obj)
         return '' if val is None else val.isoformat()
 
-    def formfield(self, **kwargs):
-        return super().formfield(**{
-            'form_class': forms.DateField,
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
+        if not form_class:
+            form_class = forms.DateField
+        return super().formfield(
+            form_class=form_class,
+            choices_form_class=choices_form_class,
             **kwargs,
-        })
+        )
 
 
 class DateTimeField(DateField):
 
-    LDAP_DATETIME_FORMATS = [
+    LDAP_DATETIME_FORMATS: List[str] = [
         "%Y%m%d%H%M%SZ",
         "%Y%m%d%H%M%S+0000"
     ]
-    LDAP_DATETIME_FORMAT = "%Y%m%d%H%M%S+0000"
+    LDAP_DATETIME_FORMAT: str = "%Y%m%d%H%M%S+0000"
 
-    empty_strings_allowed = False
-    default_error_messages = {
+    empty_strings_allowed: bool = False
+    default_error_messages: Dict[str, str] = {
         'invalid': _("'%(value)s' value has an invalid format. It must be in "
                      "YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ] format."),
         'invalid_date': _("'%(value)s' value has the correct format "
@@ -744,9 +803,9 @@ class DateTimeField(DateField):
                               "but it is an invalid date/time."),
         'invalid_ldap_datetime': _("LDAP datetime '%(value)s' value is not in a supported format"),
     }
-    description = _("Date (with time)")
+    description: str = _("Date (with time)")
 
-    def _check_fix_default_value(self):
+    def _check_fix_default_value(self) -> List[checks.Warning]:
         """
         Warn that using an actual date or datetime value is probably wrong;
         it's only evaluated on server startup.
@@ -789,7 +848,7 @@ class DateTimeField(DateField):
 
         return []
 
-    def to_python(self, value):
+    def to_python(self, value: Optional[Union[str, datetime.datetime, datetime.date]]) -> Optional[datetime.datetime]:
         if value is None:
             return value
         if isinstance(value, datetime.datetime):
@@ -803,7 +862,7 @@ class DateTimeField(DateField):
                 # call stack.
                 warnings.warn("DateTimeField %s.%s received a naive datetime "
                               "(%s) while time zone support is active." %
-                              (self.model.__name__, self.name, value),
+                              (cast(Type["Model"], self.model).__name__, self.name, value),
                               RuntimeWarning)
                 default_timezone = timezone.get_default_timezone()
                 value = timezone.make_aware(value, default_timezone)
@@ -821,9 +880,9 @@ class DateTimeField(DateField):
             )
 
         try:
-            parsed = parse_date(value)
-            if parsed is not None:
-                return datetime.datetime(parsed.year, parsed.month, parsed.day)
+            parsed_date = parse_date(value)
+            if parsed_date is not None:
+                return datetime.datetime(parsed_date.year, parsed_date.month, parsed_date.day)
         except ValueError:
             raise exceptions.ValidationError(
                 self.error_messages['invalid_date'],
@@ -837,99 +896,101 @@ class DateTimeField(DateField):
             params={'value': value},
         )
 
-    def from_db_value(self, value):
-        value = Field.from_db_value(self, value)
-        if value is []:
+    def from_db_value(self, value: List[bytes]) -> Optional[datetime.datetime]:
+        db_value = Field.from_db_value(self, value)
+        if not db_value:
             return None
-        value = value[0]
+        dt_str = db_value[0]
+        dt: Optional[datetime.datetime] = None
         for fmt in self.LDAP_DATETIME_FORMATS:
             try:
-                value = datetime.datetime.strptime(value, fmt)
+                dt = datetime.datetime.strptime(dt_str, fmt)
             except ValueError:
                 pass
             else:
                 break
-        if not isinstance(value, datetime.datetime):
+        if not isinstance(dt, datetime.datetime):
             raise exceptions.ValidationError(
                 self.error_messages['invalid_ldap_datetime'],
                 code='invalid_ldap_datetime',
                 params={'value': value},
             )
-        value = pytz.utc.localize(value)
-        return value
+        dt = pytz.utc.localize(dt)
+        return dt
 
-    def to_db_value(self, value):
+    def to_db_value(self, value: Optional[Union[datetime.datetime, datetime.date]]) -> Dict[str, List[bytes]]:
         if value:
             utc = pytz.utc
-            value = value.astimezone(utc)
-            value = value.strftime(self.LDAP_DATETIME_FORMAT)
-        return Field.to_db_value(self, value)
+            dt_str = cast(datetime.datetime, value).astimezone(utc).strftime(self.LDAP_DATETIME_FORMAT)
+        return Field.to_db_value(self, dt_str)
 
-    def pre_save(self, model_instance, add):
+    def pre_save(self, model_instance: "Model", add: bool) -> Any:
         if self.auto_now or (self.auto_now_add and add):
             value = timezone.now()
             setattr(model_instance, self.attname, value)
             return value
-        else:
-            return super().pre_save(model_instance, add)
+        return super().pre_save(model_instance, add)
 
-    def formfield(self, **kwargs):
-        return super().formfield(**{
-            'form_class': forms.DateTimeField,
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
+        if not form_class:
+            form_class = forms.DateTimeField
+        return super().formfield(
+            form_class=form_class,
+            choices_form_class=choices_form_class,
             **kwargs,
-        })
+        )
 
 
 class EmailField(CharField):
-    default_validators = [validators.validate_email]
-    description = _("Email address")
+    default_validators: List[Validator] = [dj_validators.validate_email]
+    description: str = _("Email address")
 
     def __init__(self, *args, **kwargs):
         # max_length=254 to be compliant with RFCs 3696 and 5321
         kwargs.setdefault('max_length', 254)
         super().__init__(*args, **kwargs)
 
-    def formfield(self, **kwargs):
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
         # As with CharField, this will cause email validation to be performed
         # twice.
-        return super().formfield(**{
-            'form_class': forms.EmailField,
+        if not form_class:
+            form_class = forms.EmailField
+        return super().formfield(
+            form_class=form_class,
+            choices_form_class=choices_form_class,
             **kwargs,
-        })
+        )
 
 
 class EmailForwardField(CharField):
-    default_validators = [validate_email_forward]
-    description = _("Email address")
-
-    # def __init__(self, *args, **kwargs):
-    #     # max_length=254 to be compliant with RFCs 3696 and 5321
-    #     kwargs.setdefault('max_length', 254)
-    #     super().__init__(*args, **kwargs)
-
-    # def formfield(self, **kwargs):
-    #     # As with CharField, this will cause email validation to be performed
-    #     # twice.
-    #     return super().formfield(**{
-    #         'form_class': forms.CharField,
-    #         **kwargs,
-    #     })
+    default_validators: List[Validator] = [validate_email_forward]
+    description: str = _("Email address")
 
 
 class IntegerField(Field):
-    empty_strings_allowed = False
-    default_error_messages = {
+    empty_strings_allowed: bool = False
+    default_error_messages: Dict[str, str] = {
         'invalid': _("'%(value)s' value must be an integer."),
     }
-    description = _("Integer")
+    description: str = _("Integer")  # type: ignore
 
-    def check(self, **kwargs):
+    def check(self, **kwargs) -> List[Union[checks.Error, checks.Warning]]:
         return [
             *super().check(),
             *self._check_max_length_warning(),
         ]
 
-    def _check_max_length_warning(self):
+    def _check_max_length_warning(self) -> List[checks.Warning]:
         if self.max_length is not None:
             return [
                 checks.Warning(
@@ -941,18 +1002,17 @@ class IntegerField(Field):
             ]
         return []
 
-    def from_db_value(self, value):
-        value = super().from_db_value(value)
-        if value is []:
+    def from_db_value(self, value: List[bytes]) -> Optional[int]:
+        db_value = super().from_db_value(value)
+        if not db_value:
             return None
-        return self.to_python(value[0])
+        return self.to_python(db_value[0])
 
-    def to_db_value(self, value):
-        if value:
-            value = str(value)
-        return super().to_db_value(value)
+    def to_db_value(self, value: Optional[int]) -> Dict[str, List[bytes]]:
+        db_value: Optional[str] = str(value) if value else None
+        return super().to_db_value(db_value)
 
-    def to_python(self, value):
+    def to_python(self, value: str) -> int:
         if value is None:
             return value
         try:
@@ -964,42 +1024,62 @@ class IntegerField(Field):
                 params={'value': value},
             )
 
-    def formfield(self, **kwargs):
-        return super().formfield(**{
-            'form_class': forms.IntegerField,
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
+        if not form_class:
+            form_class = forms.IntegerField
+        return super().formfield(
+            form_class=form_class,
+            choices_form_class=choices_form_class,
             **kwargs,
-        })
+        )
 
 
 class CharListField(CharField):
-    description = _("List of strings (each up to %(max_length)s)")
+    description: str = _("List of strings (each up to %(max_length)s)")
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.empty_values += ('[]', )
-        self.validators.append(validators.MaxLengthValidator(self.max_length))
+        self.validators.append(dj_validators.MaxLengthValidator(self.max_length))
 
-    def get_default(self):
+    def get_default(self) -> List[str]:
         if self.default is None:
-            return list()
-        else:
-            self._get_default()
+            return []
+        return self._get_default()
 
     def from_db_value(self, value):
+        """
+        .. note::
+
+            This is important because we don't want CharField.from_db_value() to
+            execute; we only want Field.from_db_value().  CharField.from_db_value()
+            turns the list that Field.from_db_value() returns into a string.  We
+            actually want the list.
+        """
         return Field.from_db_value(self, value)
 
-    def to_python(self, value):
+    def to_python(self, value: Optional[Union[str, List[str]]]) -> List[str]:  # type: ignore
         if not value:
             return []
-        elif isinstance(value, list):
+        if isinstance(value, list):
             return value
         return value.splitlines()
 
-    def formfield(self, **kwargs):
+    def formfield(
+        self,
+        form_class: Type[forms.Field] = None,
+        choices_form_class: Type[forms.Field] = None,
+        **kwargs
+    ) -> forms.Field:
         # Passing max_length to forms.CharField means that the value's length
         # will be validated twice. This is considered acceptable since we want
         # the value in the form field (to pass into widget for example).
-        defaults = {
+        defaults: Dict[str, Any] = {
             'max_length': self.max_length,
             'form_class': ldap_forms.CharListField,
         }
@@ -1011,26 +1091,25 @@ class CharListField(CharField):
 
 class PasswordField(CharField):
 
-    def hash_password(self, password):
-        return password
+    def hash_password(self, password: str) -> bytes:
+        return password.encode('utf-8')
 
-    def to_db_value(self, value):
+    def to_db_value(self, value: str) -> Dict[str, List[bytes]]:
         hashed_password = self.hash_password(value)
         return super().to_db_value(hashed_password)
 
 
 class LDAPPasswordField(PasswordField):
 
-    def hash_password(self, password):
+    def hash_password(self, password: str) -> bytes:
         salt = os.urandom(8)
-        h = hashlib.sha1(password)
+        h = hashlib.sha1(password.encode('utf-8'))
         h.update(salt)
-        pwhash = "{SSHA}" + encode(h.digest() + salt)
+        pwhash = "{SSHA}".encode('utf-8') + encode(h.digest() + salt)
         return pwhash
 
 
 class ADPasswordField(PasswordField):
 
-    def hash_password(self, password):
-        encoded_pass = "\"{}\"".format(password).encode("utf-16-le")
-        return encoded_pass
+    def hash_password(self, password: str) -> bytes:
+        return "\"{}\"".format(password).encode("utf-16-le")
